#ifndef PARADISO_WINDOW_HPP
#define PARADISO_WINDOW_HPP

#include "geometry.hpp"
#include "globals.hpp"

#include <functional>
#include <memory>
#include <string>

namespace paradiso {

struct Window final {
    using Size = paradiso::Size;
    using Position = paradiso::Point;

    using on_updatecallback_t = std::function<bool(
        Window&)>; //! update needs to return true to continue
    using on_resizecallback_t =
        std::function<void(Window&)>; //! resize is 'informal'
    using on_keyboardcallback_t =
        std::function<void(Window&, int, int, int, int)>;

    Window();
    ~Window();
    Window(const Window&) = delete; // no copy!
    Window(Window&&) = default;

    /**
     * @brief update internal state
     * @param cb callback to process data and render
     * @return returns false if process needs to quit
     * */
    bool update(on_updatecallback_t cb);

    Window& set_title(std::string_view title);

    Window& set_size(const Size& size);
    Size size() const;
    Size client_size() const;

    Window& set_position(const Position& position);
    Position position() const;

    bool fullscreen() const;
    Window& set_fullscreen(bool use_fullscreen);

    void set_resizecallback(on_resizecallback_t f) { on_resize_ = f; }
    void set_keyboardcallback(on_keyboardcallback_t f) { on_keyboard_ = f; }

    bool visible() const;
    Window& set_visible(bool is_visible);

  private:
    struct impl;
    std::unique_ptr<impl> impl_;

    on_resizecallback_t on_resize_;
    on_keyboardcallback_t on_keyboard_;
};

} // namespace paradiso

#endif